# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.datasets namespace.
"""

import sys as _sys

from keras.api._v2.keras.datasets import boston_housing
from keras.api._v2.keras.datasets import cifar10
from keras.api._v2.keras.datasets import cifar100
from keras.api._v2.keras.datasets import fashion_mnist
from keras.api._v2.keras.datasets import imdb
from keras.api._v2.keras.datasets import mnist
from keras.api._v2.keras.datasets import reuters